<?php

namespace Drupal\donorperfect;

use Drupal\Core\Form\FormStateInterface;

/**
 * Provides utility functions for the DonorPerfect module.
 */
class DPUtility {

  /**
   * Load the API credentials needed to access the DonorPerfect XML API.
   *
   * @param string $context
   *   The context for which the credentials are being loaded.
   *   Can be either 'settings' or 'dpquery'.
   *
   * @return array
   *   Array containing the API credentials,
   *   with possible keys 'user', 'pw', and/or 'key'.
   */
  public static function apiCredentialsLoad(string $context='settings') {
    $credentials = &drupal_static('donorperfect_api_credentials_load');
    if (!isset($credentials) || !isset($credentials[$context])) {
      $credentials[$context] = [];
      // Allow other modules to handle loading the api credentials.
      \Drupal::moduleHandler()->alter('donorperfect_api_credentials_load', $credentials[$context], $context);
      // If no other module has handled it, load credentials from default storage.
      if (empty($credentials[$context])) {
        $config = \Drupal::config('donorperfect.settings');
        foreach (['key', 'user', 'pw'] as $credential_key) {
          $credentials[$context][$credential_key] = $config->get("api.{$credential_key}");
        }
      }
    }
    return $credentials[$context];
  }

  /**
   * Validate the API credentials that were entered on the Settings form.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The $form_state from the settings form.
   */
  public static function apiCredentialsValidate(FormStateInterface $form_state) {
    // Allow other modules to handle the validation.
    $validated = FALSE;
    \Drupal::moduleHandler()->alter('donorperfect_api_credentials_validate', $form_state, $validated);
    // If no other module has handled it, perform default validation.
    if (!$validated) {
      $values = $form_state->getValues();
      if (!empty($values['api']) && is_array($values['api'])) {
        $has_credentials = FALSE;
        // If key is provided, clear user and pw.
        if (isset($values['api']['key']) && !empty(trim($values['api']['key']))) {
          $has_credentials = TRUE;
          $values['api']['key'] = trim($values['api']['key']);
          unset($values['api']['user']);
          unset($values['api']['pw']);
        } elseif (isset($values['api']['user']) && !empty(trim($values['api']['user']))) {
          $has_credentials = TRUE;
          unset($values['api']['key']);
          $values['api']['user'] = trim($values['api']['user']);
          if (isset($values['api']['pw']) && !empty(trim($values['api']['pw']))) {
            $values['api']['pw'] = trim($values['api']['pw']);
          } else {
            $form_state->setErrorByName('api][pw', t('Please enter the password that corresponds to the DonorPerfect username.'));
          }
        }
        if (!$has_credentials) {
          $form_state->setErrorByName('api][key', t('Please enter credentials to connect to the DonorPerfect XML API.'));
        }
      }
      $form_state->setValues($values);
    }
  }

  /**
   * Save the API credentials that were entered on the Settings form.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The $form_state from the settings form.
   */
  public static function apiCredentialsSave(FormStateInterface $form_state) {
    // Allow other modules to handle saving the credentials.
    $saved = FALSE;
    \Drupal::moduleHandler()->alter('donorperfect_api_credentials_save', $form_state, $saved);
    // If no other module has handled it, save the credentials in the default location.
    if (!$saved) {
      $config = \Drupal::configFactory()->getEditable('donorperfect.settings');
      $config->set('api', $form_state->getValue('api'));
      $config->save();
    }
  }

  /**
   * Load a single entity.
   *
   * @param string $entity_type_id
   * @param int $entity_id
   *
   * @return \Drupal\Core\Entity\EntityInterface|null
   */
  public static function entityLoad(string $entity_type_id, int $entity_id) {
    try {
      return \Drupal::entityTypeManager()
        ->getStorage($entity_type_id)
        ->load($entity_id);
    } catch (\Throwable $e) {
      return null;
    }
  }

  /**
   * Load multiple entities.
   *
   * @param string $entity_type_id
   * @param $entity_ids
   *   An array of entity IDs, or NULL to load all entities.
   * @param array $conditions
   * @param bool $access_check
   *
   * @return \Drupal\Core\Entity\EntityInterface[]
   */
  public static function entityLoadMultiple(string $entity_type_id, array $entity_ids=NULL, array $conditions=[], bool $access_check=TRUE) {
    try {
      if (empty($entity_ids) && !empty($conditions) && is_array($conditions)) {
        $query = \Drupal::entityQuery($entity_type_id)
          ->accessCheck($access_check);
        foreach ($conditions as $condition) {
          if (is_array($condition)) {
            if (count($condition) === 2) {
              list($field, $value) = $condition;
              $query->condition($field, $value);
            } elseif (count($condition) === 3) {
              list($field, $value, $operator) = $condition;
              $query->condition($field, $value, $operator);
            }
          }
        }
        $entity_ids = $query->execute();
      }
      return \Drupal::entityTypeManager()
        ->getStorage($entity_type_id)
        ->loadMultiple($entity_ids);
    } catch (\Throwable $e) {
      return [];
    }
  }

  /**
   * Get the specified key from the DonorPerfect cache.
   *
   * @param string $key
   *   The cache key to return.
   *
   * @return array
   */
  public static function getCache(string $key = '') {
    if (substr($key, 0, 2) !== 'dp') {
      $key = 'dp' . $key;
    }
    $cache = &drupal_static(__CLASS__ . '_getCache');
    if (!is_array($cache) || !isset($cache[$key])) {
      $config = \Drupal::config('donorperfect.cache')->get($key);
      if (!empty($config) && is_array($config)) {
        $cache[$key] = $config;
      } else {
        $cache[$key] = [];
      }
    }
    return $cache[$key];
  }

  /**
   * Return DonorPerfect codes as an array,
   *   useful for #options of a select form element.
   *
   * @param string $field_name
   *   Name of the DonorPerfect field for which the codes should be returned.
   * @param bool $include_inactive
   *   Boolean indicating whether or not inactive codes should be included.
   * @param string $include_code
   *   A DonorPerfect code that should be included even if inactive and $include_inactive is FALSE.
   *
   * @return array
   */
  public static function getCodesAsOptions(string $field_name, bool $include_inactive = FALSE, string $include_code = '') {
    $output = [];
    $dpcodes = static::getCache('codes');
    if (array_key_exists($field_name, $dpcodes) && is_array($dpcodes[$field_name])) {
      foreach ($dpcodes[$field_name] as $code) {
        if (is_array($code) && !empty($code['code']) && !empty($code['description'])) {
          if ($include_inactive || !empty($code['active']) || (!empty($include_code) && ($include_code === $code['code']))) {
            $output[$code['code']] = $code['description'];
          }
        }
      }
      asort($output);
    }
    return $output;
  }

  /**
   * Return the description for the specified DonorPerfect code.
   *
   * @param string $field_name
   *   Name of the DonorPerfect field to which the code belongs.
   * @param string $code
   *   The code for which to retrieve the description.
   *
   * @return string
   */
  public static function getCodeDescription(string $field_name, string $code) {
    $output = '';
    $codes = static::getCodesAsOptions($field_name, TRUE);
    if (array_key_exists($code, $codes)) {
      $output = $codes[$code];
    }
    return $output;
  }

  /**
   * Return DonorPerfect fields as an array,
   *   useful for #options of a select form element.
   *
   * @param string $table
   *   Name of the DonorPerfect table whose fields should be returned.
   * @param bool $prefix_table_name
   *   Boolean indicating whether or not the table name should be prefixed
   *   to the field name in the returned array keys.
   * @param bool $append_field_name
   *   Boolean indicating whether or not the field name should be appended
   *   in parentheses at the end of the returned array values.
   *
   * @return string[]
   */
  public static function getFieldsAsOptions(string $table, bool $prefix_table_name = FALSE, bool $append_field_name = TRUE) {
    $output = [];
    if (empty($table)) return $output;
    $dptables = static::getCache('tables');
    if (array_key_exists($table, $dptables) && is_array($dptables[$table])) {
      foreach ($dptables[$table] as $field_name => $field_info) {
        $key = $prefix_table_name ? "{$table}.{$field_name}" : $field_name;
        if (is_array($field_info) && !empty($field_info['prompt'])) {
          $prompt = $field_info['prompt'];
          if ($append_field_name) $prompt .= " ({$field_name})";
        } else {
          $prompt = $field_name;
        }
        $output[$key] = $prompt;
      }
      asort($output);
    }
    return $output;
  }

  /**
   * Return array of DonorPerfect fields that have been selected in the settings form
   *   to be included in the Drupal entities of specified type.
   *
   * @param string $entity_type_id
   *   The entity type for which field info should be returned.
   * @param bool $prefix_table_name
   *   Boolean indicating whether or not the table name should be prefixed
   *   to the field name in the returned array keys.
   *
   * @return array
   */
  public static function getEntityFieldInfo(string $entity_type_id, bool $prefix_table_name = FALSE) {
    $output = [];
    $entity_type = \Drupal::entityTypeManager()->getDefinition($entity_type_id);
    $base_field = $entity_type->getKey('id');
    $field_names = \Drupal::config('donorperfect.settings')->get("entity.{$entity_type_id}.fields");
    // Make sure at least the $base_field is included.
    if (empty($field_names) || !is_array($field_names) || !in_array($base_field, $field_names)) {
      if (!is_array($field_names)) $field_names = [];
      $field_names[$base_field] = $base_field;
    }
    $dptables = static::getCache('tables');
    $base_table = $entity_type->getBaseTable();
    $base_fields = (array_key_exists($base_table, $dptables) && is_array($dptables[$base_table])) ? $dptables[$base_table] : [];
    $udf_table = $entity_type->get('udf_table');
    $udf_fields = (array_key_exists($udf_table, $dptables) && is_array($dptables[$udf_table])) ? $dptables[$udf_table] : [];
    foreach ($field_names as $field_name) {
      $key = '';
      if (array_key_exists($field_name, $base_fields) && is_array($base_fields[$field_name])) {
        $key = $prefix_table_name ? "{$base_table}.{$field_name}" : $field_name;
        $field_info = $base_fields[$field_name];
      } elseif (array_key_exists($field_name, $udf_fields) && is_array($udf_fields[$field_name])) {
        $key = $prefix_table_name ? "{$udf_table}.{$field_name}" : $field_name;
        $field_info = $udf_fields[$field_name];
      } elseif ($field_name === $base_field) {
        $key = $prefix_table_name ? "{$base_table}.{$field_name}" : $field_name;
        $field_info = [
          'dp_table' => $base_table,
          'dp_field' => $base_field,
          'mysql_data_type' => 'int',
          'entity_data_type' => 'integer',
          'quote' => FALSE,
          'code' => FALSE,
          'options' => [],
          'multi' => FALSE,
          'calculated' => FALSE,
          'editable' => TRUE,
          'decimals' => FALSE,
        ];
      }
      if (!empty($key)) {
        $output[$key] = $field_info;
      }
    }
    return $output;
  }

  /**
   * Generate random string token of specified length.
   *
   * @param int $token_length_min
   *   Minimum length of token if max is also provided, otherwise the exact
   *   length of the token.
   * @param int $token_length_max
   *   Maximum length of token.
   *
   * @return string
   */
  public static function generateToken(int $token_length_min=0, int $token_length_max=0) {
    $token = '';
    if (!empty($token_length_min)) {
      if (!empty($token_length_max)) {
        if ((int)$token_length_max > (int)$token_length_min) {
          $length = rand((int)$token_length_min, (int)$token_length_max);
        } else {
          $length = rand((int)$token_length_max, (int)$token_length_min);
        }
      } else {
        $length = (int)$token_length_min;
      }
    } else {
      $length = rand(5,8);
    }
    for ($x=1; $x <= $length; $x++) {
      $char_type = rand(1,3);
      switch ($char_type) {
        case 1:
          $char = rand(48,57);
          break;
        case 2:
          $char = rand(65,90);
          break;
        case 3:
          $char = rand(97,122);
          break;
      }
      $token = $token . chr($char);
    }
    return $token;
  }

  /**
   * Validate string to ensure it contains only alpha-numerics and dashes.
   *
   * @param string $input
   *   The string to validate.
   *
   * @return boolean
   */
  public static function validateAlphaDash(string $input) {
    $input = trim($input);
    if (!empty($input)) {
      return preg_match("/^[ 0-9a-zA-Z,'#!_\-\.\/\?]+$/", $input);
    }
    return TRUE;
  }

}
