<?php

namespace Drupal\donorperfect\Plugin\views\query;

use Drupal\Core\Database\DatabaseExceptionWrapper;
use Drupal\Core\Database\Query\ConditionInterface;
use Drupal\Core\Database\Query\Select;
use Drupal\Core\Database\Query\SelectInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\views\Plugin\views\query\Sql;
use Drupal\views\Plugin\views\relationship\RelationshipPluginBase;
use Drupal\views\ResultRow;
use Drupal\views\ViewExecutable;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\donorperfect\DPQuery;
use Drupal\donorperfect\DPUtility;

/**
 * Views query plugin for DonorPerfect entities.
 *
 * @ingroup views_query_plugins
 *
 * @ViewsQuery(
 *   id = "donorperfect_views_query",
 *   title = @Translation("DonorPerfect Views Query"),
 *   help = @Translation("Query will be generated and run using the DonorPerfect API.")
 * )
 */
class DonorPerfect extends Sql {
  /**
   * An array of sections of the WHERE query. Each section is in itself
   * an array of pieces and a flag as to whether or not it should be AND
   * or OR.
   *
   * @var string
   */
  public $whereClause = '';

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('donorperfect.views.date_sql'),
      $container->get('messenger')
    );
  }

  /**
   * {@inheritdoc}
   */
  protected function defineOptions() {
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function buildOptionsForm(&$form, FormStateInterface $form_state) {
    // Some form elements belong in a fieldset for presentation, but can't
    // be moved into one because of the $form_state->getValues() hierarchy. Those
    // elements can add a #fieldset => 'fieldset_name' property, and they'll
    // be moved to their fieldset during pre_render.
    $form['#pre_render'][] = [static::class, 'preRenderAddFieldsetMarkup'];
  }

  /**
   * {@inheritdoc}
   */
  public function validateOptionsForm(&$form, FormStateInterface $form_state) {}

  /**
   * {@inheritdoc}
   */
  public function submitOptionsForm(&$form, FormStateInterface $form_state) {}

  /**
   * Generate a query and a count_query from all of the information supplied
   * to the object.
   *
   * @param $get_count
   *   Provide a count_query if this is true, otherwise provide a normal query.
   */
  public function query($get_count = FALSE) {
    $base_table = $this->view->storage->get('base_table');
    $base_field = $this->view->storage->get('base_field');
    $query = new DPQuery('select', $base_table . '.' . $base_field);
    $query->get_count = $get_count;

    // Add fields, unless this is a count query.
    if (!$get_count) {
      if (!empty($this->view->field) && is_array($this->view->field)) {
        foreach ($this->view->field as $field_def) {
          if (is_object($field_def) && !empty($field_def->table) && !empty($field_def->realField)) {
            $query->addField("{$field_def->table}.{$field_def->realField}");
            $this->fieldAliases[$field_def->table][$field_def->realField] = $field_def->field ?? $field_def->realField;
          }
        }
      }
    }

    // Add id fields for relationship tables.
    if (!empty($this->view->relationship) && is_array($this->view->relationship)) {
      foreach ($this->view->relationship as $relationship) {
        if (is_object($relationship) && ($relationship instanceof RelationshipPluginBase)) {
          if (!empty($relationship->definition) && is_array($relationship->definition)) {
            if (!empty($relationship->definition['base']) && !empty($relationship->definition['base field'])) {
              $query->addField("{$relationship->definition['base']}.{$relationship->definition['base field']}");
              $relationship_alias = $relationship->alias ?? $relationship->definition['base'];
              $this->fieldAliases[$relationship_alias][$relationship->definition['base field']] = $relationship->definition['base field'];
            }
          }
        }
      }
    }

    // Add WHERE clause.
    if (empty($this->whereClause)) {
      $this->whereClause = $this->compileWhereClause();
    }
    $query->addCondition($this->whereClause);

    // Add ORDER BY clause.
    foreach ($this->orderby as $orderby) {
      if (empty($orderby['table']) && ($pos = strpos($orderby['field'], '_'))) {
        $orderby['table'] = substr($orderby['field'], 0, $pos);
        $orderby['field'] = substr($orderby['field'], $pos + 1);
      }
      $query->addOrder($orderby['table'] . '.' . $orderby['field'] . ' ' . $orderby['direction']);
    }

    return $query;
  }

  /**
   * Compile conditions into a string to be used for the WHERE clause
   *   of the SQL statement.
   *
   * @return string
   */
  protected function compileWhereClause() {
    $whereClause = '';
    $hasWhere = FALSE;

    // First check to see if there are any where conditions.
    if (is_array($this->where)) {
      foreach ($this->where as $index => $value) {
        if (isset($value['conditions']) && !empty($value['conditions'])) {
          $hasWhere = TRUE;
          break;
        }
      }
    }

    if ($hasWhere) {
      // Replace relationship aliases with actual table names.
      if (!empty($this->view->relationship) && is_array($this->view->relationship)) {
        $find = $replace = [];
        foreach ($this->view->relationship as $relationship) {
          if (is_object($relationship) && ($relationship instanceof RelationshipPluginBase)) {
            if (!empty($relationship->definition) && is_array($relationship->definition)) {
              if (!empty($relationship->definition['base']) && !empty($relationship->alias)) {
                $find[] = $relationship->alias;
                $replace[] = $relationship->definition['base'];
              }
            }
          }
        }
        $this->conditionsReplaceRelationshipAliases($this->where, $find, $replace);
      }

      $condition = $this->buildCondition('where');
      $connection = $this->getConnection();
      $queryPlaceholder = new Select($connection, 'node');
      $condition->compile($connection, $queryPlaceholder);

      $whereClause = $condition->__toString();
      $args = $condition->arguments();
      $counter = 0;
      $needsEscapeClause = FALSE;

      $this->quoteArgs($args, $condition->conditions(), $counter, $needsEscapeClause);
      $this->mergePlaceholders($whereClause, $args);

      // Replace the any ESCAPE clauses with a single backslash.
      if ($needsEscapeClause) {
        $whereClause = str_replace("ESCAPE '\\\\'","ESCAPE '\\'",$whereClause);
      } else {
        $whereClause = str_replace(" ESCAPE '\\\\'","",$whereClause);
      }

      // Clean up the $whereClause to try to save characters in the final API call.
      $whereClause = str_replace('"', '', $whereClause);
      $whereClause = str_replace("  ", " ", $whereClause);
      $whereClause = str_replace("( (", "((", $whereClause);
      $whereClause = str_replace(") )", "))", $whereClause);
      $whereClause = str_replace("AND(", "AND (", $whereClause);
      $whereClause = str_replace(")AND", ") AND", $whereClause);
      $whereClause = str_replace("OR(", "OR (", $whereClause);
      $whereClause = str_replace(")OR", ") OR", $whereClause);
    }

    return $whereClause;
  }

  /**
   * If the view has relationships, replace relationship aliases in conditions
   * with actual table names.
   *
   * @param array $conditions
   *   Array of conditions to search for relationship aliases.
   * @param array $find
   *   Array of relationship aliases to be replaced.
   * @param array $replace
   *   Array of actual table names corresponding to the relationship aliases.
   */
  protected function conditionsReplaceRelationshipAliases(&$conditions, $find=[], $replace=[]) {
    if (is_array($conditions)) {
      foreach ($conditions as $key => $condition) {
        if (is_array($condition)) {
          if (!empty($condition['field'])) {
            $conditions[$key]['field'] = str_replace($find, $replace, $condition['field']);
          }
          if (!empty($condition['conditions']) && is_array($condition['conditions'])) {
            // Recursively call this function.
            $this->conditionsReplaceRelationshipAliases($conditions[$key]['conditions'], $find, $replace);
          }
        }
      }
    }
  }


  /**
   * This function is modeled after DatabaseCondition::compile in order to match up 
   * the argument placeholders and determine whether or not the arguments need 
   * to be quoted in the final query.
   */
  protected function quoteArgs(&$args, $conditions, &$counter, &$needsEscapeClause) {
    foreach ($conditions as $condition) {
      if (empty($condition['operator'])) {
        continue;
      } else {
        if ($condition['field'] instanceof ConditionInterface) {
          // Recursively call quoteArgs.
          $this->quoteArgs($args, $condition['field']->conditions(), $counter, $needsEscapeClause);
        } else {
          $use_value = TRUE;
          if (in_array(strtoupper($condition['operator']), array('IS NULL', 'IS NOT NULL'))) $use_value = FALSE;
          if ($condition['value'] instanceof SelectInterface) {
            // Recursively call quoteArgs.
            $this->quoteArgs($args, $condition['value']->conditions(), $counter, $needsEscapeClause);
            $use_value = FALSE;
          } elseif (!is_array($condition['value'])) {
            $condition['value'] = [$condition['value']];
          }

          if ($use_value && is_array($condition['value'])) {
            $thisQuote = FALSE;
            if (is_string($condition['field'])) {
              $field = $condition['field'];
              $thisDot = strpos($field, '.');
              if ($thisDot !== FALSE) {
                $thisTable = trim(substr($field, 0, $thisDot));
                $thisField = trim(substr($field, $thisDot + 1, strlen($field) - 1));
                $dptables = DPUtility::getCache('tables');

                if (array_key_exists($thisTable, $dptables) && is_array($dptables[$thisTable]) && array_key_exists($thisField, $dptables[$thisTable]) && is_array($dptables[$thisTable][$thisField])) 
                {
                  // Determine if this field should be quoted in a query.
                  if (isset($dptables[$thisTable][$thisField]['quote'])) {
                    $thisQuote = (bool)$dptables[$thisTable][$thisField]['quote'];
                  }
                }
              }
            }

            foreach ($condition['value'] as $value) {
              $placeholder = ':db_condition_placeholder_' . $counter;
              $counter++;
              // Add quotation marks where appropriate, and also escape any single quotes already contained in $value.
              if ($thisQuote) {
                $args[$placeholder] = "'" . str_replace("'", "''", $value) . "'";
              } else {
                $args[$placeholder] = str_replace("'", "''", $value);
              }

              if (strpos($value, '\\') !== FALSE) $needsEscapeClause = TRUE;
            }
          }
        }
      }
    }
  }

  /**
   * Merge placeholder values into the where clause.
   *
   * @param string $whereClause
   *   The where clause that contains placeholders.
   * @param array $args
   *   The placeholder values as prepared by quoteArgs().
   */
  protected function mergePlaceholders(string &$whereClause, array &$args) {
      // Update the query with the new placeholders.
      // preg_replace is necessary to ensure the replacement does not affect
      // placeholders that start with the same exact text. For example, if the
      // query contains the placeholders :foo and :foobar, and :foo has an
      // array of values, using str_replace would affect both placeholders,
      // but using the following preg_replace would only affect :foo because
      // it is followed by a non-word character.
    foreach ($args as $key => $value) {
      $whereClause = preg_replace('#' . $key . '\b#', $value, $whereClause);
    }
  }


  /**
   * Executes the query and fills the associated view object with according
   * values.
   *
   * Values to set: $view->result, $view->total_rows, $view->execute_time,
   * $view->current_page.
   */
  public function execute(ViewExecutable $view) {
    $external = FALSE; // Whether this query will run against an external database.
    $query = $view->build_info['query'];
    $count_query = $view->build_info['count_query'];

    $items = [];
    if (!empty($query)) {
      $start = microtime(TRUE);

      try {
        if ($view->pager->useCountQuery() || !empty($view->get_total_rows)) {
          $view->pager->executeCountQuery($count_query);
        }

        // Let the pager modify the query to add limits.
        $view->pager->preExecute($query);

        if (!empty($this->limit) || !empty($this->offset)) {
          // We can't have an offset without a limit, so provide a very large limit instead.
          $limit  = intval(!empty($this->limit) ? $this->limit : 999999);
          $offset = intval(!empty($this->offset) ? $this->offset : 0);
          $query->queryRange($offset, $limit);
        }

        $results = $query->execute()->result;
        $view->result = [];
        $index = 0;
        foreach ($results as $result) {
          $obj = new ResultRow();
          if (!empty($view->field) && is_array($view->field)) {
            foreach ($view->field as $name => $def) {
              if (!is_object($def) || empty($def->realField)) continue;
              $field_alias = !empty($def->field_alias) && ($def->field_alias !== 'unknown') ? $def->field_alias : '';
              if (empty($field_alias)) $field_alias = $def->field ?? '';
              if (empty($field_alias)) continue;
              if (isset($result[$def->realField])) {
                $obj->{$field_alias} = $result[$def->realField];
              } else {
                $obj->{$field_alias} = NULL;
              }
            }
          }
          if (!empty($view->relationship) && is_array($view->relationship)) {
            foreach ($view->relationship as $relationship) {
              if (is_object($relationship) && ($relationship instanceof RelationshipPluginBase)) {
                if (!empty($relationship->definition) && is_array($relationship->definition)) {
                  if (!empty($relationship->definition['base']) && !empty($relationship->definition['base field'])) {
                    $relationship_alias = $relationship->alias ?? $relationship->definition['base'];
                    $field_alias = $this->getFieldAlias($relationship_alias, $relationship->definition['base field']) ?: $relationship->definition['base field'];
                    if (isset($result[$relationship->definition['base field']])) {
                      $obj->{$field_alias} = $result[$relationship->definition['base field']];
                    }
                  }
                }
              }
            }
          }
          $obj->index = $index;
          $index++;
          $view->result[] = $obj;
        }

        $view->pager->postExecute($view->result);
        $view->pager->updatePageInfo();
        $view->total_rows = $view->pager->getTotalItems();

        // Load all entities contained in the results.
        $this->loadEntities($view->result);
      }
      catch (\Throwable $e) {
        $view->result = [];
        if (!empty($view->live_preview)) {
          $this->messenger->addError($e->getMessage());
        } else {
          throw new DatabaseExceptionWrapper("Exception in {$view->storage->label()}[{$view->storage->id()}]: {$e->getMessage()}");
        }
      }
    } else {
      $start = microtime(TRUE);
    }
    $view->execute_time = microtime(TRUE) - $start;
  }

}
